class MdException(Exception):
    pass


class NotLoggedIn(MdException):
    pass


class ActionForbidden(MdException):
    pass


class RefreshTokenFailed(MdException):
    pass


class DownloadException(MdException):
    pass


class NoFollowRedirect(DownloadException):
    pass


class InvalidStatusCode(DownloadException):
    pass


class InvalidFileLength(DownloadException):
    pass



from typing import List, Optional

from pydantic.decorator import validate_arguments

from ..util import PaginatedRequest
from ..endpoints import Endpoints
from ..schema import Type, TypeOrId, User, ScanlationGroup
from .base import APIBase


class GroupAPI(APIBase):
    @validate_arguments
    def search(
        self,
        name: str = None,
        ids: List[TypeOrId[ScanlationGroup]] = None,
        limit: int = 10,
        offset: int = 0,
    ):
        return PaginatedRequest(self.api, Endpoints.Group.SEARCH, params={
            "name": name, "ids": ids
        }, limit=limit, offset=offset)

    @validate_arguments
    def create(
        self,
        name: str,
        leader: TypeOrId[User],
        members: List[TypeOrId[User]]
    ) -> ScanlationGroup:
        return Type.parse_obj(self.api._make_request(
            Endpoints.Group.CREATE, body={
                "name": name,
                "leader": leader,
                "members": members,
            }
        ))

    @validate_arguments
    def get(self, group: TypeOrId[ScanlationGroup]) -> ScanlationGroup:
        return Type.parse_obj(self.api._make_request(
            Endpoints.Group.GET, urlparams={
                "group": group
            }
        ))

    @validate_arguments
    def edit(
        self,
        group: ScanlationGroup,
        name: str,
        locked: bool,
        leader: TypeOrId[User],
        members: List[TypeOrId[User]],
    ) -> ScanlationGroup:
        return Type.parse_obj(self.api._make_request(
            Endpoints.Group.EDIT, body={
                "name": name,
                "locked": locked,
                "leader": leader,
                "version": group.version,
                "members": members
            }, urlparams={
                "group": group.id
            }
        ))

    @validate_arguments
    def delete(self, group: TypeOrId[ScanlationGroup]):
        self.api._make_request(
            Endpoints.Group.DELETE, urlparams={"group": group}
        )

    @validate_arguments
    def follow(self, group: TypeOrId[ScanlationGroup]):
        self.api._make_request(
            Endpoints.Group.FOLLOW, urlparams={"group": group}
        )

    @validate_arguments
    def unfollow(self, group: TypeOrId[ScanlationGroup]):
        self.api._make_request(
            Endpoints.Group.UNFOLLOW, urlparams={"group": group}
        )



from enum import Enum
import functools
import base64
import json
import time
from typing import Generic, Iterable, List, TypeVar
from queue import Queue, Empty
from threading import Thread

from pydantic.main import BaseModel

from .schema import Type, UnsetValue


def _type_id(type):
    raise NotImplementedError


def _get_token_expires(jwt):
    payload = jwt.split(".")[1]
    payload = json.loads(base64.b64decode(payload + "=="))
    return payload["exp"]


def _is_token_expired(jwt):
    return _get_token_expires(jwt) <= time.time()


def shadows(hoc):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            kwargs.update(zip(func.__code__.co_varnames, args))
            return hoc(**kwargs)
        return wrapper
    return decorator


def params_to_query(kwargs):
    query = {}
    for k, v in kwargs.items():
        if isinstance(v, BaseModel):
            v = v.dict()
        if isinstance(v, list):
            query[f"{k}[]"] = v
        elif isinstance(v, dict):
            for k2, v2 in v.items():
                if isinstance(v2, (list, dict)):
                    raise ValueError("Deep structures not supported")
                if isinstance(v2, Enum):
                    query[f"{k}[{k2}]"] = v2.value
                elif v2 is UnsetValue:
                    query[f"{k}[{k2}]"] = None
                elif v2 is not None:
                    query[f"{k}[{k2}]"] = v2
        elif isinstance(v, Enum):
            query[k] = v.value
        elif v is UnsetValue:
            query[k] = None
        elif v is not None:
            query[k] = v
    return query


def strip_nulls(object):
    return {
        (k.value if isinstance(k, Enum) else k): (
            None if v is UnsetValue
            else v.value if isinstance(v, Enum)
            else strip_nulls(v) if isinstance(v, dict)
            else v
        )
        for k, v in (object or {}).items()
        if v is not None
    }


class Worker:
    class Kill:
        pass
    KILL = Kill()

    def __init__(self, worker, long_lived=False, num_workers=3):
        self._worker = worker
        self._long_lived = long_lived
        self._num_workers = num_workers
        self._queue = Queue()
        self._killed = False

    @property
    def killed(self):
        return self._killed

    def _local_worker(self):
        while True:
            if self._long_lived:
                job = self._queue.get()
            else:
                try:
                    job = self._queue.get_nowait()
                except Empty:
                    return
            if job is self.KILL:
                self._queue.task_done()
                return

            try:
                self._worker(job)
            finally:
                self._queue.task_done()

    def start(self):
        for _ in range(self._num_workers):
            Thread(target=self._local_worker, daemon=True).start()

    def kill(self):
        self._killed = True
        while True:
            try:
                self._queue.get_nowait()
            except Empty:
                break
            else:
                self._queue.task_done()
        for _ in range(self._num_workers):
            self._queue.put(self.KILL)

    def enqueue(self, job):
        if self._killed:
            return
        self._queue.put(job)

    def join(self):
        self._queue.join()


T = TypeVar("T")


class PaginatedRequest(Generic[T]):
    _LIMIT = 10

    def __init__(
        self, api, *args, limit=None, offset=None, params=None, **kwargs
    ):
        self.total = None
        self.offset = offset if offset is not None else 0
        self._limit = limit if limit is not None else self._LIMIT
        self.has_more = True

        self._results = None
        self._params = params or {}
        self._api = api
        self._args = args
        self._kwargs = kwargs

        self._ensure_populated()

    def _get_next(self) -> None:
        results = self._api._make_request(*self._args, **self._kwargs, params={
            **self._params,
            "offset": self.offset,
            "limit": self._limit
        })
        self.total = results.get("total", 0)
        self.offset += results.get("limit", 0)
        self._results = results.get("results", [])
        self.has_more = self.offset < self.total

    def __iter__(self) -> Iterable[T]:
        return self

    def _ensure_populated(self) -> None:
        if self._results is None or len(self._results) == 0:
            if self.total is not None and self.offset >= self.total:
                raise StopIteration
            self._get_next()

    def __len__(self) -> int:
        self._ensure_populated()
        return self.total

    def __next__(self) -> T:
        self._ensure_populated()

        if len(self._results) == 0:
            raise StopIteration

        result = self._results.pop(0)
        if "relationships" in result:
            result["data"]["relationships"] = result["relationships"]
        result = result.get("data", result)
        return Type.parse_obj(result)

    def next_page(self) -> List[T]:
        try:
            self._ensure_populated()
        except StopIteration:
            return []
        res = self._results
        self._results = []
        res = [i.get("data", i) for i in res]
        return [Type.parse_obj(i) for i in res]


__all__ = (
    "_type_id", "_get_token_expires", "_is_token_expired", "PaginatedRequest"
)



from ..endpoints import Endpoints
from ..schema import Type, User
from .base import APIBase


class AccountAPI(APIBase):
    """
    Actions related to Managdex accounts. These actions generally relate
    to the ``/account`` endpoints.
    """

    def create(self, username: str, password: str, email: str) -> User:
        """
        Create a new Mangadex account.

        :param username: New account's username
        :param password: New account's password
        :param email: New account's email

        :returns: The newly created user
        """
        return Type.parse_obj(self.api._make_request(
            Endpoints.Account.CREATE, body={
                "username": username,
                "password": password,
                "email": email
            }, auth=False
        ))

    def recover(self, email: str) -> None:
        """
        Begin an account recovery. This will send an email to ``email``
        if an account exists with that email address.

        :param email: The account's email address
        """
        self.api._make_request(Endpoints.Account.RECOVER, body={
            "email": email
        }, auth=False)

    def complete_recover(self, code: str, password: str) -> None:
        """
        Complete an account recovery.

        :param code: The code received via email
        :param password: The new password to change to
        """
        self.api._make_request(Endpoints.Account.COMPLETE_RECOVER, body={
            "newPassword": password
        }, urlparams={
            "code": code
        }, auth=False)

    def activate(self, code: str) -> None:
        """
        Activate a new Mangadex account.

        :param code: The code received via email
        """
        self.api._make_request(Endpoints.Account.ACTIVATE, urlparams={
            "code": code
        })

    def activate_resend(self, email: str) -> None:
        """
        Request a new email be sent during the account activate process.

        :param email: The new user's email address
        """
        self.api._make_request(Endpoints.Account.ACTIVATE_RESEND, body={
            "email": email
        })



from typing import List

from pydantic import validate_arguments

from ..util import PaginatedRequest
from ..endpoints import Endpoints
from ..schema import TypeOrId, Type, CustomListVisibility, Manga, CustomList
from .base import APIBase


class ListAPI(APIBase):
    @validate_arguments
    def create(
        self,
        name: str,
        visibility: CustomListVisibility,
        manga: List[TypeOrId[Manga]]
    ) -> CustomList:
        return Type.parse_obj(self.api._make_request(
            Endpoints.List.CREATE,
            body={
                "name": name,
                "visibility": visibility,
                "manga": manga,
                "version": 1,
            }
        ))

    @validate_arguments
    def edit(
        self,
        custom_list: CustomList,
        name: str,
        visibility: CustomListVisibility,
        manga: List[TypeOrId[Manga]]
    ):
        self.api._make_request(
            Endpoints.List.EDIT, body={
                "name": name,
                "visibility": visibility,
                "manga": manga,
                "version": custom_list.version,
            }, urlparams={"list": custom_list.id}
        )

    @validate_arguments
    def get(self, list_id: TypeOrId[CustomList]) -> CustomList:
        return Type.parse_obj(self.api._make_request(
            Endpoints.List.GET, urlparams={
                "list": list_id
            }
        ))

    @validate_arguments
    def delete(self, list_id: TypeOrId[CustomList]) -> None:
        self.api._make_request(
            Endpoints.List.DELETE, urlparams={
                "list": list_id
            }
        )

    @validate_arguments
    def get_feed(
        self,
        list_id: TypeOrId[CustomList],
        limit: int = 10,
        offset: int = 0,
    ):
        return PaginatedRequest(
            self.api,
            Endpoints.List.GET_FEED,
            urlparams={"list": list_id},
            limit=limit, offset=offset
        )



from typing import List, BinaryIO

from pydantic.decorator import validate_arguments

from ..util import PaginatedRequest
from ..endpoints import Endpoints
from ..schema import (
    Author, TypeOrId, CanUnset, CoverSortOrder, Cover, Manga, User, Type
)
from .base import APIBase


class CoverAPI(APIBase):
    @validate_arguments
    def search(
        self,
        manga: List[TypeOrId[Manga]] = None,
        ids: List[TypeOrId[Cover]] = None,
        uploaders: List[TypeOrId[User]] = None,
        order: CoverSortOrder = None,
        limit: int = 10,
        offset: int = 0,
    ) -> PaginatedRequest[Cover]:
        return PaginatedRequest(self.api, Endpoints.Cover.SEARCH, params={
            "manga": manga, "ids": ids, "uploaders": uploaders, "order": order
        }, limit=limit, offset=offset)

    def upload(
        self, manga: TypeOrId[Manga], file_: BinaryIO
    ) -> Cover:
        manga = TypeOrId.return_type(manga)
        return Type.parse_obj(self.api._make_request(
            Endpoints.Cover.UPLOAD, urlparams={"manga": manga},
            files={"file": file_}
        ))

    @validate_arguments
    def edit(
        self, cover: Cover, volume: CanUnset[str] = None,
        description: CanUnset[str] = None
    ) -> Cover:
        body = {"version": cover.version}
        if volume is not None:
            body["volume"] = volume
        if description is not None:
            body["description"] = volume
        return Type.parse_obj(self.api._make_request(
            Endpoints.Cover.EDIT, body=body, urlparams={"cover": cover.id}
        ))

    @validate_arguments
    def delete(self, cover: TypeOrId[Author]) -> None:
        self.api._make_request(
            Endpoints.Cover.DELETE, urlparams={"cover": cover}
        )



from uuid import UUID
from mdapi.schema.search import ChapterSortOrder
from typing import Dict, List
from datetime import datetime

from pydantic import validate_arguments

from ..util import PaginatedRequest, shadows
from ..endpoints import Endpoints
from ..schema import (
    MultiMode, Status, LanguageCode, PublicationDemographic, Type, Tag, Year,
    ContentRating, TypeOrId, MangaSortOrder, ReadingStatus, Author, Manga,
    LinksKey, Version, CustomList, LocalizedString, CanUnset
)
from .base import APIBase


class MangaAPI(APIBase):
    def _search(self, limit=None, offset=None, **kwargs):
        return PaginatedRequest(
            self.api, Endpoints.Manga.SEARCH, params=kwargs,
            limit=limit, offset=offset,
        )

    @validate_arguments
    @shadows(_search)
    def search(
        self,
        title: str = None,
        authors: List[TypeOrId[Author]] = None,
        artists: List[TypeOrId] = None,
        year: Year = None,
        includedTags: List[TypeOrId[Tag]] = None,
        includedTagsMode: MultiMode = None,
        excludedTags: List[TypeOrId[Tag]] = None,
        excludedTagsMode: MultiMode = None,
        status: Status = None,
        originalLanguage: LanguageCode = None,
        publicationDemographic: PublicationDemographic = None,
        ids: List[TypeOrId[Manga]] = None,
        contentRating: ContentRating = None,
        createdAtSince: datetime = None,
        updatedAtSince: datetime = None,
        order: MangaSortOrder = None,
        limit: int = 10,
        offset: int = 0,
    ):
        """
        :param title:
        :param authors:
        :param artists:
        :param year:
        :param includedTags:
        :param includedTagsMode:
        :param excludedTags:
        :param excludedTagsMode:
        :param status:
        :param originalLanguage:
        :param publicationDemographic:
        :param ids:
        :param contentRating:
        :param createdAtSince:
        :param updatedAtSince:
        :param order:
        :param limit:
        :param offset:
        """
        ...

    @validate_arguments
    def get(self, manga: TypeOrId[Manga]) -> Manga:
        return Type.parse_obj(self.api._make_request(
            Endpoints.Manga.GET,
            urlparams={"manga": manga}
        ))

    @validate_arguments
    def delete(self, manga: TypeOrId[Manga]) -> None:
        self.api._make_request(
            Endpoints.Manga.DELETE,
            urlparams={"manga": manga}
        )

    @validate_arguments
    def follow(self, manga: TypeOrId[Manga]) -> None:
        self.api._make_request(
            Endpoints.Manga.FOLLOW,
            urlparams={"manga": manga}
        )

    @validate_arguments
    def unfollow(self, manga: TypeOrId[Manga]) -> None:
        self.api._make_request(
            Endpoints.Manga.UNFOLLOW,
            urlparams={"manga": manga}
        )

    def all_tags(self) -> List[Tag]:
        return [
            Type.parse_obj(i.get("data"))
            for i in self.api._make_request(Endpoints.Manga.TAGS)
        ]

    @validate_arguments
    def get_batch_read(self, ids: List[TypeOrId[Manga]]) -> List[UUID]:
        if len(ids) == 0:
            return []
        return self.api._make_request(Endpoints.Manga.BATCH_GET_READ, params={
            "ids": ids
        })

    def random(self) -> Manga:
        return Type.parse_obj(self.api._make_request(Endpoints.Manga.RANDOM))

    def _create(self, **kwargs):
        kwargs["version"] = 1
        return self.api._make_request(Endpoints.Manga.CREATE, kwargs)

    @validate_arguments
    @shadows(_create)
    def create(
        self,
        title: LocalizedString,
        altTitles: List[LocalizedString],
        description: LocalizedString,
        authors: List[TypeOrId[Author]],
        artists: List[TypeOrId],
        links: List[Dict[LinksKey, str]],
        originalLanguage: LanguageCode,
        year: Year,
        contentRating: ContentRating,
        lastVolume: str = None,
        lastChapter: str = None,
        publicationDemographic: PublicationDemographic = None,
        status: Status = None,
    ) -> Manga:
        ...

    def _edit(self, **kwargs):
        manga = kwargs.pop("manga")
        return self.api._make_request(
            Endpoints.Manga.EDIT, kwargs,
            urlparams={"manga": manga}
        )

    @validate_arguments
    @shadows(_edit)
    def edit(
        self,
        manga: TypeOrId[Manga],
        title: LocalizedString,
        altTitles: List[LocalizedString],
        description: LocalizedString,
        authors: List[TypeOrId[Author]],
        artists: List[TypeOrId],
        links: Dict[LinksKey, str],
        originalLanguage: LanguageCode,
        version: Version,
        year: CanUnset[Year] = None,
        lastVolume: CanUnset[str] = None,
        lastChapter: CanUnset[str] = None,
        publicationDemographic: CanUnset[PublicationDemographic] = None,
        status: CanUnset[Status] = None,
        contentRating: CanUnset[ContentRating] = None,
    ) -> Manga:
        ...

    def _get_chapters(self, limit=None, offset=None, **kwargs):
        manga = kwargs.pop("manga")
        return PaginatedRequest(
            self.api, Endpoints.Manga.CHAPTERS, params=kwargs,
            urlparams={"manga": manga}, limit=limit, offset=offset,
        )

    @validate_arguments
    @shadows(_get_chapters)
    def get_chapters(
        self,
        manga: TypeOrId[Manga],
        translatedLanguage: List[LanguageCode] = None,
        createdAtSince: datetime = None,
        updatedAtSince: datetime = None,
        publishAtSince: datetime = None,
        order: ChapterSortOrder = None,
        limit: int = 10,
        offset: int = 0,
    ) -> PaginatedRequest:
        ...

    @validate_arguments
    def get_read(self, manga: TypeOrId[Manga]):
        return self.api._make_request(
            Endpoints.Manga.MARK_READ,
            urlparams={"manga": manga}
        )

    @validate_arguments
    def set_status(self, manga: TypeOrId[Manga], status: ReadingStatus):
        return self.api._make_request(
            Endpoints.Manga.SET_STATUS,
            urlparams={"manga": manga},
            body={"status": status}
        )

    @validate_arguments
    def add_to_list(self, manga: TypeOrId[Manga], list: TypeOrId[CustomList]):
        self.api._make_request(
            Endpoints.Manga.ADD_TO_LIST,
            urlparams={"manga": manga, "list": list}
        )

    @validate_arguments
    def remove_from_list(
        self, manga: TypeOrId[Manga], list: TypeOrId[CustomList]
    ):
        self.api._make_request(
            Endpoints.Manga.REMOVE_FROM_LIST,
            urlparams={"manga": manga, "list": list}
        )



class APIBase:
    """
    The base class for all API classes.
    """

    def __init__(self, md, api):
        # If we import at the root level we have a circular import!
        from ..mdapi import APIHandler, MdAPI

        self.md: MdAPI = md
        self.api: APIHandler = api



from typing import List
from pydantic import validate_arguments

from ..endpoints import Endpoints
from ..schema import TypeOrId, Chapter, LegacyType, MappingID, Type
from .base import APIBase


class MiscAPI(APIBase):
    @validate_arguments
    def get_md_at_home_url(
        self, chapter: TypeOrId[Chapter], force_port_443: bool = False
    ):
        return self.api._make_request(Endpoints.GET_MD_AT_HOME, urlparams={
            "chapter": chapter
        }, params={
            "forcePort443": True if force_port_443 else None
        })["baseUrl"]

    @validate_arguments
    def solve_captcha(self, challenge: str):
        self.api._make_request(Endpoints.SOLVE_CAPTCHA, body={
            "captchaChallenge": challenge
        })

    @validate_arguments
    def legacy_mapping(
        self, manga_ids: List[int], type: LegacyType = "manga"
    ) -> List[MappingID]:
        return [
            Type.parse_obj(i.get("data", i))
            for i in self.api._make_request(Endpoints.LEGACY_MAPPING, body={
                "type": type,
                "ids": manga_ids
            })
        ]

    @validate_arguments
    def report_mdah(
        self, url: str, success: bool, cached: bool, num_bytes: int,
        duration: int
    ):
        return
        self.api._make_request(Endpoints.MDAH_REPORT, body={
            "url": url, "success": success, "cached": cached,
            "bytes": num_bytes, "duration": duration
        })



class Endpoints:
    COVERS = "https://uploads.mangadex.org/covers/{}/{}"

    class Manga:
        SEARCH = ("GET", "/manga")
        CREATE = ("POST", "/manga")
        TAGS = ("GET", "/manga/tag")
        BATCH_GET_READ = ("GET", "/manga/read")
        RANDOM = ("GET", "/manga/random")
        GET = ("GET", "/manga/{manga}")
        EDIT = ("PUT", "/manga/{manga}")
        DELETE = ("DELETE", "/manga/{manga}")

        SET_STATUS = ("POST", "/manga/{manga}/status")

        FOLLOW = ("POST", "/manga/{manga}/follow")
        UNFOLLOW = ("DELETE", "/manga/{manga}/follow")

        CHAPTERS = ("GET", "/manga/{manga}/feed")
        MARK_READ = ("GET", "/manga/{manga}/read")

        ADD_TO_LIST = ("POST", "/manga/{manga}/list/{list}")
        REMOVE_FROM_LIST = ("DELETE", "/manga/{manga}/list/{list}")

    class Cover:
        SEARCH = ("GET", "/cover")
        UPLOAD = ("POST", "/cover/{manga}")
        EDIT = ("PUT", "/cover/{cover}")
        DELETE = ("DELETE", "/cover/{cover}")

    class Auth:
        LOGIN = ("POST", "/auth/login")
        CHECK = ("GET", "/auth/check")
        LOGOUT = ("POST", "/auth/logout")
        REFRESH = ("POST", "/auth/refresh")

    class Account:
        CREATE = ("POST", "/account/create")
        RECOVER = ("POST", "/account/recover")
        COMPLETE_RECOVER = ("POST", "/account/recover/{code}")
        ACTIVATE = ("GET", "/account/activate/{code}")
        ACTIVATE_RESEND = ("POST", "/account/activate/resend")

    class Group:
        SEARCH = ("GET", "/group")
        CREATE = ("POST", "/group")
        GET = ("GET", "/group/{group}")
        EDIT = ("PUT", "/group/{group}")
        DELETE = ("DELETE", "/group/{group}")
        FOLLOW = ("POST", "/group/{group}/follow")
        UNFOLLOW = ("DELETE", "/group/{group}/follow")

    class List:
        CREATE = ("POST", "/list")
        GET = ("GET", "/list/{list}")
        EDIT = ("PUT", "/list/{list}")
        DELETE = ("DELETE", "/list/{list}")
        GET_FEED = ("GET", "/list/{list}/feed")

    class User:
        GET_ME = ("GET", "/user/me")
        LIST = ("GET", "/user/list")
        OTHER_LIST = ("GET", "/user/{user}/list")
        FOLLOWS_GROUP = ("GET", "/user/follows/group")
        FOLLOWS_MANGA = ("GET", "/user/follows/manga")
        FOLLOWS_CHAPTERS = ("GET", "/user/follows/manga/feed")

    class Chapter:
        SEARCH = ("GET", "/chapter")
        GET = ("GET", "/chapter/{chapter}")
        EDIT = ("PUT", "/chapter/{chapter}")
        DELETE = ("DELETE", "/chapter/{chapter}")
        MARK_READ = ("POST", "/chapter/{chapter}/read")
        MARK_UNREAD = ("DELETE", "/chapter/{chapter}/read")

    class Author:
        CREATE = ("POST", "/author")
        SEARCH = ("GET", "/author")
        GET = ("GET", "/author/{author}")
        EDIT = ("PUT", "/author/{author}")
        DELETE = ("DELETE", "/author/{author}")

    class Upload:
        GET_SESSION = ("GET", "/upload")
        BEGIN = ("GET", "/upload/begin")
        ADD_IMAGE = ("POST", "/upload/{session}")
        DELETE_IMAGE = ("DELETE", "/upload/{session}/{file}")
        ABADON = ("DELETE", "/upload/{session}")
        COMMIT = ("POST", "/upload/{session}/commit")

    GET_MD_AT_HOME = ("GET", "/at-home/server/{chapter}")
    SOLVE_CAPTCHA = ("POST", "/captcha/solve")
    LEGACY_MAPPING = ("POST", "/legacy/mapping")

    MDAH_REPORT = ("POST", "https://api.mangadex.network/report")



import click
import os
import platform
import json

import requests

from .exceptions import (
    MdException, NotLoggedIn, ActionForbidden, RefreshTokenFailed
)
from .api import (
    AccountAPI, AuthAPI, AuthorAPI, ChapterAPI, GroupAPI, ListAPI, MangaAPI,
    MiscAPI, UserAPI, CoverAPI, UploadAPI
)
from .util import _is_token_expired, params_to_query, strip_nulls
from .endpoints import Endpoints


class APIHandler:
    UA = f"PyMyAPI on Python {platform.python_version()}"
    BASE = "https://api.mangadex.org"
    BASE = "https://api.mangadex.org.dev.mdcloud.moe"

    AUTH_FILE = ".mdauth"

    DEBUG = False

    def __init__(self, md):
        self.md = md
        self.user = None
        self._auth = None

        self.captcha = None

    def _save_auth(self):
        with open(self.AUTH_FILE, "w") as auth_file:
            json.dump({
                "user": self.user,
                "_auth": self._auth
            }, auth_file)

    def _load_auth(self):
        if not os.path.exists(self.AUTH_FILE):
            return

        with open(self.AUTH_FILE) as auth_file:
            try:
                auth = json.load(auth_file)
            except json.JSONDecodeError:
                return

        try:
            user = auth["user"]
            _auth = auth["_auth"]
        except KeyError:
            return

        self.user = user
        self._auth = _auth

        self._check_expired(silent_error=True)

    def _check_expired(self, silent_error=False):
        if self._auth is None:
            return

        if _is_token_expired(self._auth["session"]):
            try:
                self.md.auth.refresh()
            except MdException:
                if silent_error:
                    self.user = self._auth = None
                else:
                    raise RefreshTokenFailed()

    def _get_headers(self, auth: bool = True) -> dict:
        headers = {}
        if auth and self._auth is not None:
            headers["Authorization"] = "Bearer " + self._auth.get("session")
        if self.captcha is not None:
            headers["X-Captcha-Result"] = self.captcha
        headers["User-Agent"] = self.UA
        return headers

    def _make_request(
        self, action, body=None, params=None, urlparams=None, auth=True,
        files=None
    ):
        if params is not None:
            params = params_to_query(params)
        if action != Endpoints.Auth.REFRESH:
            self._check_expired()

        needs_base = not action[1].startswith(("http://", "https://"))
        url = (
            (self.BASE if needs_base else "")
            + action[1].format(**(urlparams or {}))
        )
        req = requests.request(
            action[0], url,
            json=None if action[0] == "GET" else strip_nulls(body),
            files=files,
            params=strip_nulls(params),
            headers=self._get_headers(auth)
        )
        if self.DEBUG:
            click.echo(click.style(f" -> {action[0]} {req.url}", fg="yellow"))
            from devtools import debug
            if action[0] != "GET":
                debug(strip_nulls(body))

            correlation = req.headers.get("X-Correlation-ID")
            if correlation:
                click.echo(click.style(
                    f" :: Correlation: {correlation}", fg="yellow"
                ))

        try:
            resp = {} if req.status_code == 204 else req.json()
        except json.decoder.JSONDecodeError:
            resp = None

        if req.status_code == 401:
            raise NotLoggedIn(resp)
        elif req.status_code == 403:
            raise ActionForbidden(resp)

        if req.status_code < 200 or req.status_code > 299:
            if resp is None:
                raise MdException(req.status_code)
            raise MdException(resp.get("errors", []))

        if not isinstance(resp, dict):
            return resp
        if "data" in resp:
            data = resp["data"]
            if "relationships" in resp:
                data["relationships"] = resp["relationships"]
            resp = data
        return resp

    def _authenticate(self, username, token):
        self._auth = token
        if token is None or username is not None:
            self.user = {"username": username}
        self._save_auth()

    def _get_refresh_token(self):
        if self._auth:
            return self._auth["refresh"]
        return None


class MdAPI:
    DEBUG = False

    def __init__(self):
        self.api = APIHandler(self)
        self.api.DEBUG = self.DEBUG

        self.account = AccountAPI(self, self.api)
        self.auth = AuthAPI(self, self.api)
        self.author = AuthorAPI(self, self.api)
        self.chapter = ChapterAPI(self, self.api)
        self.cover = CoverAPI(self, self.api)
        self.group = GroupAPI(self, self.api)
        self.list = ListAPI(self, self.api)
        self.manga = MangaAPI(self, self.api)
        self.misc = MiscAPI(self, self.api)
        self.user = UserAPI(self, self.api)
        self.upload = UploadAPI(self, self.api)

        self.api._load_auth()



from pydantic.decorator import validate_arguments

from ..endpoints import Endpoints
from ..exceptions import MdException
from .base import APIBase


class AuthAPI(APIBase):
    """
    Actions related to Managdex authentication. The seperation between
    these actions and :mod:`mdapi.api.account` are to remain in-line
    with the structure of the upstream API. These actions generally
    relate to the ``/account`` endpoints.
    """

    @validate_arguments
    def login(self, username: str, password: str) -> None:
        """
        Login to a Mangadex account

        :param username: Account's username
        :param password: Account's password
        """
        try:
            token = self.api._make_request(Endpoints.Auth.LOGIN, {
                "username": username,
                "password": password
            }).get("token")
        except MdException:
            raise

        self.api._authenticate(username, token)

    def check(self) -> bool:
        """
        Check if a user is currently logged in

        :returns: If a user is currently logged in
        """
        return self.api._make_request(Endpoints.Auth.CHECK)

    def logout(self) -> None:
        """
        Logout the current user. If no user is logged in, this will
        still make a logout request, but it will have no effect.
        """
        try:
            self.api._make_request(Endpoints.Auth.LOGOUT)
        except MdException:
            pass
        self.api._authenticate(None, None)

    def refresh(self) -> None:
        """
        Refresh the current session token using the refresh token.
        """
        ref = self.api._get_refresh_token()
        if ref is None:
            raise MdException("Not logged in")

        try:
            token = self.api._make_request(
                Endpoints.Auth.REFRESH, {"token": ref}
            ).get("token")
        except MdException:
            raise

        self.api._authenticate(None, token)



from typing import List, BinaryIO
from pydantic.decorator import validate_arguments

from ..endpoints import Endpoints
from ..schema import (
    TypeOrId, Type, Manga, ScanlationGroup, UploadSession, UploadSessionFile
)
from ..util import _type_id
from .base import APIBase


class UploadAPI(APIBase):
    def get_session(self):
        return Type.parse_obj(
            self.api._make_request(Endpoints.Upload.GET_SESSION)
        )

    @validate_arguments
    def begin(
        self,
        manga: TypeOrId[Manga],
        groups: List[TypeOrId[ScanlationGroup]]
    ):
        return Type.parse_obj(
            self.api._make_request(Endpoints.Upload.BEGIN, body={
                "manga": manga, "groups": groups
            })
        )

    def upload_images(
        self,
        session: TypeOrId[UploadSession],
        files: List[BinaryIO]
    ):
        files = {f"file{n + 1}": i for n, i in enumerate(files)}
        return [
            Type.parse_obj(i.get("data"))
            for i in self.api._make_request(
                Endpoints.Upload.ADD_IMAGE.format(session=_type_id(session)),
                files=files
            )
        ]

    @validate_arguments
    def delete_image(
        self,
        session: TypeOrId[UploadSession],
        file: TypeOrId[UploadSessionFile]
    ):
        self.api._make_request(
            Endpoints.Upload.DELETE_IMAGE.format(session=session, file=file)
        )

    @validate_arguments
    def abandon(self, session: TypeOrId[UploadSession]):
        self.api._make_request(Endpoints.Upload.ABADON.format(session=session))

    @validate_arguments
    def commit(
        self,
        session: TypeOrId[UploadSession],
        page_order: TypeOrId[UploadSessionFile],
        volume: str,
        chapter: str,
        title: str,
        translated_language: str
    ):
        return Type.parse_obj(self.api._make_request(
            Endpoints.Upload.COMMIT.format(session=session),
            body={
                "chapterDraft": {
                    "volume": volume,
                    "chapter": chapter,
                    "title": title,
                    "translatedLanguage": translated_language,
                },
                "pageOrder": page_order
            }
        ))



import os
from typing import BinaryIO, Generator, List, Tuple
from datetime import datetime

from pydantic.decorator import validate_arguments
import requests

from ..util import PaginatedRequest, shadows
from ..endpoints import Endpoints
from ..exceptions import (
    DownloadException, NoFollowRedirect, InvalidStatusCode, InvalidFileLength
)
from ..schema import (
    ChapterSortOrder, LanguageCode, Type, Manga, TypeOrId, Chapter,
    ScanlationGroup, User, CanUnset
)
from .base import APIBase


class ChapterAPI(APIBase):
    """
    Actions related to Managdex chapters. These actions generally relate
    to the ``/chapter`` endpoints.
    """

    def _search(self, limit=None, offset=None, **kwargs):
        return PaginatedRequest(
            self.api, Endpoints.Chapter.SEARCH, params=kwargs,
            limit=limit, offset=offset,
        )

    @validate_arguments
    @shadows(_search)
    def search(
        self,
        title: str = None,
        ids: List[TypeOrId[Chapter]] = None,
        groups: List[TypeOrId[ScanlationGroup]] = None,
        uploader: TypeOrId[User] = None,
        manga: TypeOrId[Manga] = None,
        volume: str = None,
        chapter: str = None,
        translatedLanguage: LanguageCode = None,
        createdAtSince: datetime = None,
        updatedAtSince: datetime = None,
        publishAtSince: datetime = None,
        order: ChapterSortOrder = None,
        limit: int = 10,
        offset: int = 0,
    ):
        """
        Search for a manga.

        :param title: The chapter title to search
        :param ids: A whitelist of chapter IDs to search
        :param groups: A whitelist of groups to search chapters from
        :param uploader: The chapter's uploader
        :param manga: Manga to request chapters from
        :param volume: The volume this chapter is in
        :param chapter: The chapter number
        :param translatedLanguage: The language this was translated into
        :param createdAtSince: Only show chapters created after this time
        :param updatedAtSince: Only show chapters updated after this time
        :param publishAtSince: Only show chapters publisged after this time
        :param order: The search order
        :param limit: The number of results per page
        :param offset: The offset to start from
        """
        ...

    @validate_arguments
    def get(self, chapter: TypeOrId[Chapter]) -> Chapter:
        """
        Request full details for an chapter. This notably includes image
        filenames.

        :param chapter: The author to lookup. Either an
            `mdapi.schema.Chapter` object, or its UUID.

        :returns: The chapter, if found
        """
        return Type.parse_obj(self.api._make_request(
            Endpoints.Chapter.GET,
            urlparams={"chapter": chapter}
        ))

    def _edit(self, **kwargs):
        chapter = kwargs.pop("chapter")
        kwargs["chapter"] = kwargs.pop("chapterNumber")
        return self.api._make_request(
            Endpoints.Chapter.EDIT, kwargs,
            urlparams={"chapter": chapter}
        )

    @validate_arguments
    @shadows(_edit)
    def edit(
        self,
        chapter: Chapter,
        title: str,
        volume: CanUnset[str] = None,
        chapterNumber: CanUnset[str] = None,
        translatedLanguage: CanUnset[LanguageCode] = None,
    ) -> Chapter:
        """
        Exit a chapter. This notably includes image
        filenames.

        :param chapter: As a this makes use of version numbers for
            conflict resolution, a `mdapi.schema.Chapter` object is
            required rather than optional.
        :param title: The new title for the chapter
        :param volume: The new volume number
        :param chapterNumber: The new chapter number
        :param translatedLanguage: This chapter's translated language
        """
        ...

    @validate_arguments
    def delete(self, chapter: TypeOrId[Chapter]) -> None:
        """
        Delete a chapter.

        :param chapter: The chapter to delete. Either an
            `mdapi.schema.Chapter` object, or its UUID.
        """
        self.api._make_request(Endpoints.Chapter.DELETE, urlparams={
            "chapter": chapter
        })

    @validate_arguments
    def mark_read(self, chapter: TypeOrId[Chapter]) -> None:
        """
        Mark a chapter as read.

        :param chapter: The chapter to mark as read. Either an
            `mdapi.schema.Chapter` object, or its UUID.
        """
        self.api._make_request(Endpoints.Chapter.MARK_READ, urlparams={
            "chapter": chapter
        })

    @validate_arguments
    def mark_unread(self, chapter: TypeOrId[Chapter]) -> None:
        """
        Mark a chapter as unread.

        :param chapter: The chapter to mark as unread. Either an
            `mdapi.schema.Chapter` object, or its UUID.
        """
        self.api._make_request(Endpoints.Chapter.MARK_UNREAD, urlparams={
            "chapter": chapter
        })

    @validate_arguments
    def page_urls_for(
        self, chapter: Chapter, data_saver: bool = False
    ) -> Generator[str, None, None]:
        """
        Get a list of all page image URLs for this chapter.

        .. note::
            If the ``MD_CHAPTER_BASE_URL`` environment variable is set,
            the API is not queried for an MD@H server, and the set value
            is instead used. This could be useful when testing MD@H
            clients.

        :param chater: The chapter to get the pages for
        :param data_saver: Should data-saver URLs be provided instead?

        :returns: A generator that produces page URLs
        """
        # Manual override for testing
        if "MD_CHAPTER_BASE_URL" in os.environ:
            base = os.environ.get("MD_CHAPTER_BASE_URL")
        else:
            base = self.api.md.misc.get_md_at_home_url(chapter)

        base += "/data-saver/" if data_saver else "/data/"
        base += chapter.hash + "/"

        for i in chapter.dataSaver if data_saver else chapter.data:
            yield base + i

    @validate_arguments
    def download_page(
        self, url: str, follow_redirect: bool = False, report_mdah: bool = True
    ) -> Generator[Tuple[bytes, int], None, None]:
        """
        Download a single page of a manga.

        :param url: The URL for the page
        :param follow_redirect: Should redirects be followed?
        :param report_mdah: Report node statistics to MD@H.

        .. note::
            Unless there is a good reason, it is recommended to always
            leave ``report_mdah`` set to ``True``, to ensure faulty
            nodes are accurately monitored.

        :returns: A generator that yields ``(chunk, total_length)``.
        """
        try:
            req = requests.get(url, stream=True)
        except requests.RequestException:
            if report_mdah:
                self.md.misc.report_mdah(url, False, False, 0, 0)
            raise DownloadException

        if req.url != url and not follow_redirect:
            if report_mdah:
                self.md.misc.report_mdah(
                    url, False, False, 0, req.elapsed.microseconds // 1000
                )
            raise NoFollowRedirect(req.url)

        if req.status_code != 200:
            if report_mdah:
                self.md.misc.report_mdah(
                    url, False, False, 0, req.elapsed.microseconds // 1000
                )

            raise InvalidStatusCode(req.status_code, req.content)

        is_cached = req.headers.get("X-Cache", "").startswith("HIT")
        total_length = int(req.headers.get("Content-Length"))

        if not total_length:
            if report_mdah:
                self.md.misc.report_mdah(
                    url, False, is_cached, 0, req.elapsed.microseconds // 1000
                )

            raise InvalidFileLength(total_length)

        chunk_size = 131072  # 0.125 MB
        bytes_downloaded = 0

        try:
            for chunk in req.iter_content(chunk_size=chunk_size):
                bytes_downloaded += len(chunk)
                yield (chunk, total_length)
        except requests.RequestException:
            if report_mdah:
                self.md.misc.report_mdah(
                    url, False, is_cached, bytes_downloaded,
                    req.elapsed.microseconds // 1000
                )
            raise DownloadException()

        if report_mdah:
            self.md.misc.report_mdah(
                url, True, is_cached, bytes_downloaded,
                req.elapsed.microseconds // 1000
            )

    def download_page_to(
        self,
        url: str, output: BinaryIO, follow_redirect: bool = False,
        is_iter: bool = False, report_mdah: bool = True
    ) -> Generator[Tuple[int, int], None, None]:
        """
        Download a page into a file-like. Paramaters are mostly the same
        as `mdapi.chapter.Chapter.download_page`.

        :param url: The URL for the page
        :param output: The output file-like to save to
        :param follow_redirect: Should redirects be followed?
        :param is_iter: Should this function return a generator?
        :param report_mdah: Report node statistics to MD@H.

        :returns: A generator yielding ``(downloaded, total_length)`` if
            ``is_iter`` is ``True``.
        """
        downloaded = 0
        for chunk, total_length in (
            self.download_page(url, follow_redirect, report_mdah)
        ):
            downloaded += len(chunk)
            output.write(chunk)
            if is_iter:
                yield (downloaded, total_length)



from typing import List

from pydantic.decorator import validate_arguments

from ..util import PaginatedRequest
from ..endpoints import Endpoints
from ..schema import Author, Type, TypeOrId, CanUnset, AuthorSortOrder
from .base import APIBase


class AuthorAPI(APIBase):
    """
    Actions related to authors. These actions generally relate to the
    ``/author`` endpoints.
    """

    @validate_arguments
    def create(self, name: str) -> Author:
        """
        Create a new author.

        :param name: The name for the author

        :returns: The newly created author
        """
        return Type.parse_obj(self.api._make_request(
            Endpoints.Author.CREATE,
            body={"name": name}
        ))

    @validate_arguments
    def get(self, author: TypeOrId[Author]) -> Author:
        """
        Request full details for an author.

        :param author: The author to lookup. Either an
            `mdapi.schema.Author` object, or their UUID.

        :returns: The author, if found
        """
        return Type.parse_obj(self.api._make_request(
            Endpoints.Author.GET,
            urlparams={"author": author}
        ))

    @validate_arguments
    def search(
        self,
        name: str = None,
        ids: List[TypeOrId[Author]] = None,
        order: AuthorSortOrder = None,
        limit: int = 10,
        offset: int = 0,
    ) -> PaginatedRequest[Author]:
        """
        Search for an author by name.

        :param name: The name to search with
        :param ids: Whitelist of authors to search from
        :param order: The order to sort results
        :param limit: The number of results per page
        :param offset: The offset to start from

        :returns: Paginated search results
        """
        return PaginatedRequest(self.api, Endpoints.Author.SEARCH, params={
            "name": name, "ids": ids, "order": order
        }, limit=limit, offset=offset)

    @validate_arguments
    def edit(
        self, author: Author, name: CanUnset[str] = None
    ) -> None:
        """
        Make edits to a user. Currently only changing the author name is
        supported.

        :param author: The author to edit. Either an
            `mdapi.schema.Author` object, or their UUID.
        :param name: The new name for the author
        """

        body = {"version": author.version}
        if name is not None:
            body["name"] = name
        self.api._make_request(
            Endpoints.Author.EDIT, body=body, urlparams={"author": author.id}
        )

    @validate_arguments
    def delete(self, author: TypeOrId[Author]) -> None:
        """
        Delete an author.

        :param author: The author to delete. Either an
            `mdapi.schema.Author` object, or their UUID.
        """
        self.api._make_request(
            Endpoints.Author.delete, urlparams={"author": author}
        )



from pydantic.decorator import validate_arguments

from ..endpoints import Endpoints
from ..util import PaginatedRequest
from ..schema import TypeOrId, User, Type
from .base import APIBase


class UserAPI(APIBase):
    @validate_arguments
    def get_list(self, limit: int = 10, offset: int = 0):
        return PaginatedRequest(
            self.api, Endpoints.User.LIST, limit=limit, offset=offset
        )

    @validate_arguments
    def get_list_for(
        self, user: TypeOrId[User], limit: int = 10, offset: int = 0
    ):
        return PaginatedRequest(
            self.api,
            Endpoints.User.OTHER_LIST,
            urlparams={
                "user": user
            },
            limit=limit, offset=offset
        )

    def get_self(self) -> User:
        return Type.parse_obj(self.api._make_request(Endpoints.User.GET_ME))

    @validate_arguments
    def get_followed_groups(self, limit: int = 10, offset: int = 0):
        return PaginatedRequest(
            self.api, Endpoints.User.FOLLOWS_GROUP, limit=limit, offset=offset
        )

    @validate_arguments
    def get_followed_chapters(self, limit: int = 10, offset: int = 0):
        return PaginatedRequest(
            self.api, Endpoints.User.FOLLOWS_CHAPTERS,
            limit=limit, offset=offset
        )

    @validate_arguments
    def get_followed_manga(self, limit: int = 10, offset: int = 0):
        return PaginatedRequest(
            self.api, Endpoints.User.FOLLOWS_MANGA, limit=limit, offset=offset
        )



import setuptools


with open("README.md", encoding="utf-8") as file_:
    long_description = file_.read()


setuptools.setup(
    name="mdex",
    version="0.0.14",
    author="Bottersnike",
    author_email="bottersnike237@gmail.com",
    description="Mangadex CLI",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Bottersnike/mdapi",
    project_urls={
        "Bug Tracker": "https://github.com/Bottersnike/mdapi/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
    ],
    scripts=[
        "scripts/mdex.py"
    ],
    packages=["mdapi", "mdapi.api"],
    python_requires=">=3.8",
    install_requires=open("requirements.txt").read().split("\n")
)



import functools
from mdapi.schema.const import SortOrder
from mdapi.schema.search import ChapterSortOrder
import os

import click

from .mdapi import MdAPI, MdException, NotLoggedIn


def sanitize(x):
    return "".join(i for i in x if (i.isalnum() or i in "._- "))


def uses_md(func):
    @click.option("--debug", is_flag=True)
    @functools.wraps(func)
    def wrapper(debug, *args, **kwargs):
        MdAPI.DEBUG = debug
        md = MdAPI()
        return func(md, *args, **kwargs)
    return wrapper


@click.group()
def cli():
    pass


@cli.command()
@uses_md
def login(md: MdAPI):
    if md.user is not None:
        if not click.confirm(
            "You are already logged in. "
            "Would you like to login to a different account?"
        ):
            return
        md.auth.logout()

    username = click.prompt("Username")
    password = click.prompt("Password", hide_input=True)

    try:
        md.auth.login(username, password)
    except NotLoggedIn:
        click.echo(click.style("Username or password incorrect!", fg="red"))
    except MdException:
        click.echo(click.style("Failed to login.", fg="red"))
    else:
        click.echo(click.style(f"Logged in as {username}", fg="green"))


@cli.command()
@uses_md
def logout(md: MdAPI):
    if md.user is None:
        click.echo(click.style("You are not logged in", fg="red"))
    else:
        md.auth.logout()


@cli.command()
@uses_md
def whoami(md: MdAPI):
    try:
        user = md.user.get_self()
    except NotLoggedIn:
        click.echo(click.style("Not logged in", fg="red"))
        return

    click.echo(user.username)


@cli.command()
@uses_md
@click.argument("query", nargs=-1)
def search(md: MdAPI, query):
    results = md.manga.search(title=" ".join(query))
    results._ensure_populated()

    click.echo(click.style(f" -=- {results.total} results -=-", fg="green"))

    while (page := results.next_page()):
        for i in page:
            click.echo(click.style(i.id, fg="magenta"), nl=False)
            click.echo(" ", nl=False)
            click.echo(click.style(str(i.title), fg="bright_blue"))
        if not results.has_more:
            break
        if not click.confirm("Show more?"):
            break


@cli.command()
@uses_md
@click.argument("manga", nargs=1)
@click.option("-l", "--locales", default="en")
def chapters(md: MdAPI, manga, locales):
    results = md.manga.get_chapters(
        manga, translatedLanguage=locales.split(","),
        order=ChapterSortOrder(chapter=SortOrder.desc)
    )
    results._ensure_populated()

    click.echo(click.style(f" -=- {results.total} chapters -=-", fg="green"))

    while (page := results.next_page()):
        for i in page:
            click.echo(click.style(i.id, fg="magenta"), nl=False)
            click.echo(" ", nl=False)
            click.echo(
                click.style(f"({i.chapter}) ", fg="bright_blue"), nl=False
            )
            if i.title:
                click.echo(click.style(str(i.title), fg="blue"), nl=False)
            click.echo("")
        if not results.has_more:
            break
        if not click.confirm("Show more?"):
            break


def download_chapter(md: MdAPI, chapter, path):
    for n, page in enumerate(md.chapter.page_urls_for(chapter)):
        ext = page.split(".")[-1]
        filename = f"{n + 1:03}.{ext}"
        with open(os.path.join(path, filename), "wb") as f:
            with click.progressbar(label=filename, length=1) as bar:
                bar.update(0)
                for downloaded, total_length in (
                    md.chapter.download_page_to(page, f, is_iter=True)
                ):
                    bar.pos = downloaded / total_length
                    bar.update(0)
                bar.pos = 1
                bar.update(0)


def read_manga(manga, locales=("en", )):
    md = MdAPI()
    manga = md.manga.get(manga)
    results = md.manga.get_chapters(manga.id, translatedLanguage=locales)
    results._ensure_populated()
    if not click.confirm(
        f"This will download {results.total} chapters. Proceed?"
    ):
        return

    for chapter in results:
        click.echo(
            click.style(f"Downloading chapter {chapter.chapter}", fg="green")
        )
        path = (
            f"Manga/{sanitize(str(manga.title) or 'No title')}/"
            f"{chapter.translatedLanguage}-{chapter.chapter}/"
        )
        click.echo(click.style(f"Downloading to {path}", fg="green"))
        os.makedirs(path, exist_ok=True)
        download_chapter(md, chapter, path)


@cli.command()
@uses_md
@click.argument("chapter", nargs=1)
@click.option("-l", "--locales", default="en")
def read(md: MdAPI, chapter, locales):
    try:
        chapter_ = md.chapter.get(chapter)
    except MdException:
        chapter_ = None

    if chapter_ is None:
        return read_manga(chapter, locales.split(","))
    chapter = chapter_

    manga = None

    for i in chapter.relationships:
        if i.type == "manga":
            manga = md.manga.get(i.id)
            break
    else:
        click.echo(click.style("Failed to locate parent manga", fg="red"))
        return

    path = (
        f"Manga/{sanitize(str(manga.title) or 'No title')}/{chapter.chapter}/"
    )
    click.echo(click.style(f"Downloading to {path}", fg="green"))
    os.makedirs(path, exist_ok=True)

    download_chapter(md, chapter, path)


@cli.command()
@uses_md
@click.argument("email", nargs=1)
def recover(md: MdAPI, email):
    md.account.recover(email)
    click.echo(click.style("Check your inbox for an email!", fg="green"))


@cli.command()
@uses_md
@click.argument("code", nargs=1)
def complete_recover(md: MdAPI, code):
    password = click.prompt("Password", hide_input=True)
    md.account.complete_recover(code, password)
    click.echo(click.style("Password reset", fg="green"))


@cli.command()
@uses_md
def follows(md: MdAPI):
    try:
        for follow in md.user.get_followed_manga():
            click.echo(click.style(follow.id, fg="magenta"), nl=False)
            click.echo(" ", nl=False)
            click.echo(click.style(str(follow.title), fg="bright_blue"))
    except NotLoggedIn:
        click.echo(click.style("Not logged in", fg="red"))
        return


def main():
    cli()



from .account import AccountAPI
from .auth import AuthAPI
from .author import AuthorAPI
from .chapter import ChapterAPI
from .cover import CoverAPI
from .group import GroupAPI
from .list import ListAPI
from .manga import MangaAPI
from .misc import MiscAPI
from .user import UserAPI
from .upload import UploadAPI


__all__ = (
    "AccountAPI", "AuthAPI", "AuthorAPI", "ChapterAPI", "GroupAPI", "ListAPI",
    "MangaAPI", "MiscAPI", "UserAPI", "CoverAPI", "UploadAPI"
)
